@@ -22,29 +22,42 @@
 
 #include "config.h"
 
+#include "gkr-ask-tool.h"
+#include "gkr-ask-request.h"
+
+#include <gtk/gtk.h>
+#include <glib/gi18n.h>
+
 #include <stdio.h>
 #include <string.h>
 #include <locale.h>
 #include <stdlib.h>
 #include <errno.h>
 #include <sys/mman.h>
+#include <unistd.h>
 
-#include <gtk/gtk.h>
-#include <glib/gi18n.h>
+static GKeyFile *input_data = NULL;
+static GKeyFile *output_data = NULL;
+static gboolean grabbed = FALSE;
 
-#include "gkr-ask-request.h"
-
-static const gchar *env_title = NULL;
-static const gchar *env_primary = NULL;
-static const gchar *env_secondary = NULL;
-static guint env_flags = 0;
+/* Because Solaris doesn't have err() :( */
+static void 
+fatal (const char *msg1, const char *msg2)
+{
+	fprintf (stderr, "%s: %s%s%s\n", 
+	         g_get_prgname (),
+	         msg1 ? msg1 : "", 
+	         msg1 && msg2 ? ": " : "",
+	         msg2 ? msg2 : "");
+	exit (1);
+}
 
 static gchar*
 create_markup (const gchar *primary, const gchar *secondary)
 {
 	/* We're passed markup for both of these */
 	return g_strdup_printf ("<span weight=\"bold\" size=\"larger\">%s</span>\n\n%s", 
-	                        primary, secondary);
+	                        primary, secondary ? secondary : "");
 }
 
 static gchar*
@@ -130,7 +143,7 @@ unlock_memory (void)
 static void
 lock_memory (void)
 {
-	int r = -1;
+	int r = 0;
 
 	/* 
 	 * TODO: This is a copout, due to the fact that GTK, and the entry 
@@ -149,11 +162,43 @@ lock_memory (void)
 		g_atexit (unlock_memory);
 }
 
+static gboolean
+grab_keyboard (GtkWidget *win, GdkEvent *event, gpointer data)
+{
+	if (!grabbed)
+		if (gdk_keyboard_grab (win->window, FALSE, gdk_event_get_time (event)))
+			g_message ("could not grab keyboard");
+	grabbed = TRUE;
+	return FALSE;
+}
+
+static gboolean
+ungrab_keyboard (GtkWidget *win, GdkEvent *event, gpointer data)
+{
+	if (!grabbed)
+		gdk_keyboard_ungrab (gdk_event_get_time (event));
+	grabbed = FALSE;
+	return FALSE;
+}
+
+static gboolean
+window_state_changed (GtkWidget *win, GdkEventWindowState *event, gpointer data)
+{
+	GdkWindowState state = gdk_window_get_state (win->window);
+	
+	if (state & GDK_WINDOW_STATE_WITHDRAWN ||
+	    state & GDK_WINDOW_STATE_ICONIFIED ||
+	    state & GDK_WINDOW_STATE_FULLSCREEN ||
+	    state & GDK_WINDOW_STATE_MAXIMIZED)
+	    	ungrab_keyboard (win, (GdkEvent*)event, data);
+	else
+		grab_keyboard (win, (GdkEvent*)event, data);
+		
+	return FALSE;
+}
+
 static gint
-run_dialog (const char *title,
-	    const char *primary,
-	    const char *secondary,
-	    gboolean include_password,
+run_dialog (gboolean include_password,
 	    gboolean include_confirm,
 	    gboolean include_original,
 	    char **password_out,
@@ -165,22 +210,27 @@ run_dialog (const char *title,
 	GtkWidget *dialog;
 	GtkLabel *message_widget;
 	GtkLabel *notice;
-	char *message;
+	char *message, *value, *value2;
 	char *notice_text;
 	GtkWidget *old, *entry, *confirm;
-	GtkWidget *vbox;
+	GtkWidget *vbox, *label;
 	gint response;
 	va_list args;
 	const char *text;
 	gint response_id;
 	GtkWidget *table, *ptable;
-	GtkWidget *image;
+	GtkWidget *image, *check, *location;
 	const char *password;
 	const char *confirmation;
 	const char *original;
 	int row;
 
-	dialog = gtk_dialog_new_with_buttons (title , NULL, 0, NULL, NULL);
+	value = g_key_file_get_value (input_data, "general", "title", NULL);
+	if (!value)
+		fatal ("no 'title' field in input data", NULL);
+	dialog = gtk_dialog_new_with_buttons (value, NULL, 0, NULL, NULL);
+	g_free (value);
+	
 	gtk_window_set_icon_name(GTK_WINDOW(dialog), "stock_lock");
 	gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
  	gtk_container_set_border_width (GTK_CONTAINER (dialog), 6);
@@ -223,7 +273,15 @@ run_dialog (const char *title,
 	gtk_table_attach (GTK_TABLE (table), image, 
 	                  0, 1, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
 	
-	message = create_markup (primary, secondary);
+	value = g_key_file_get_value (input_data, "general", "primary", NULL);
+	if (!value)
+		fatal ("no 'primary' field in input data", NULL);
+	value2 = g_key_file_get_value (input_data, "general", "secondary", NULL);
+
+	message = create_markup (value, value2);
+	g_free (value);
+	g_free (value2);
+	
 	message_widget = GTK_LABEL (gtk_label_new (message));
 	g_free (message);
 	gtk_label_set_use_markup (message_widget, TRUE);
@@ -249,77 +307,83 @@ run_dialog (const char *title,
 	gtk_table_set_col_spacings (GTK_TABLE (ptable), 12);
 	gtk_container_set_border_width (GTK_CONTAINER (ptable), 5);
 	gtk_box_pack_start (GTK_BOX (vbox), ptable, FALSE, TRUE, 0);
-	
+
+	location = gkr_ask_tool_create_location (input_data);
+	if (location) {
+		gtk_table_resize (GTK_TABLE (ptable), ++row, 2);
+		label = gtk_label_new_with_mnemonic (_("_Location:"));
+		gtk_label_set_mnemonic_widget (GTK_LABEL (label), location);
+		gtk_table_attach (GTK_TABLE (ptable), label,
+				  0, 1, row - 1, row,
+				  GTK_FILL, GTK_SHRINK, 0, 0);
+		gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
+		gtk_table_attach_defaults (GTK_TABLE (ptable), location,
+		                           1, 2, row - 1, row);
+	}
+		
 	old = NULL;
 	if (include_original) {
-		GtkWidget *label_old;
-	
-		label_old = gtk_label_new_with_mnemonic (_("_Old password:"));
+		gtk_table_resize (GTK_TABLE (ptable), ++row, 2);	
+		label = gtk_label_new_with_mnemonic (_("_Old password:"));
 		old = gtk_entry_new ();
 		gtk_entry_set_visibility (GTK_ENTRY (old), FALSE);
-		gtk_label_set_mnemonic_widget (GTK_LABEL (label_old), 
-					       old);
+		gtk_label_set_mnemonic_widget (GTK_LABEL (label), old);
 		g_signal_connect_swapped (old,
 					  "activate",
 					  G_CALLBACK (gtk_window_activate_default),
 					  dialog);
-		gtk_table_attach (GTK_TABLE (ptable), label_old,
-				  0, 1, row, row+1,
+		gtk_table_attach (GTK_TABLE (ptable), label,
+				  0, 1, row - 1, row,
 				  GTK_FILL, GTK_SHRINK, 0, 6);
-		gtk_misc_set_alignment (GTK_MISC (label_old), 0.0, 0.5);
+		gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
 		gtk_table_attach (GTK_TABLE (ptable), old,
-				  1, 2, row, row+1, 
+				  1, 2, row - 1, row, 
 				  GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 6);
-		row++;
 	}
 	
 	entry = NULL;
 	if (include_password) {
-		GtkWidget *label_entry;
-		
-		label_entry = gtk_label_new_with_mnemonic (_("_Password:"));
+		gtk_table_resize (GTK_TABLE (ptable), ++row, 2);	
+		label = gtk_label_new_with_mnemonic (_("_Password:"));
 		entry = gtk_entry_new ();
 		gtk_entry_set_visibility (GTK_ENTRY (entry), FALSE);
-		gtk_label_set_mnemonic_widget (GTK_LABEL (label_entry), 
-					       entry);
+		gtk_label_set_mnemonic_widget (GTK_LABEL (label), entry);
 		g_signal_connect_swapped (entry,
 					  "activate",
 					  G_CALLBACK (gtk_window_activate_default),
 					  dialog);
-		gtk_table_attach (GTK_TABLE (ptable), label_entry,
-				  0, 1, row, row+1,
+		gtk_table_attach (GTK_TABLE (ptable), label,
+				  0, 1, row - 1, row,
 				  GTK_FILL, GTK_SHRINK, 0, 0);
-		gtk_misc_set_alignment (GTK_MISC (label_entry), 0.0, 0.5);
+		gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
 		gtk_table_attach_defaults (GTK_TABLE (ptable), 
 					   entry,
-					   1, 2, row, row+1);
-		row++;
+					   1, 2, row - 1, row);
 	}
 
 	confirm = NULL;
 	if (include_confirm) {
-		GtkWidget *label_confirm;
 		GtkWidget *strength_bar;
 	
-		gtk_table_resize (GTK_TABLE (ptable),4,2);
-		label_confirm = gtk_label_new_with_mnemonic (_("_Confirm password:"));
+		gtk_table_resize (GTK_TABLE (ptable), ++row, 2);	
+		label = gtk_label_new_with_mnemonic (_("_Confirm password:"));
 		confirm = gtk_entry_new ();
 		gtk_entry_set_visibility (GTK_ENTRY (confirm), FALSE);
-		gtk_label_set_mnemonic_widget (GTK_LABEL (label_confirm), confirm);
+		gtk_label_set_mnemonic_widget (GTK_LABEL (label), confirm);
 		g_signal_connect_swapped (confirm,
 					  "activate",
 					  G_CALLBACK (gtk_window_activate_default),
 					  dialog);
-		gtk_table_attach (GTK_TABLE (ptable), label_confirm,
-				  0, 1, row, row+1,
+		gtk_table_attach (GTK_TABLE (ptable), label,
+				  0, 1, row - 1, row,
 				  GTK_FILL, GTK_SHRINK, 0, 0);
-		gtk_misc_set_alignment (GTK_MISC (label_confirm), 0.0, 0.5);
+		gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
 		gtk_table_attach_defaults (GTK_TABLE (ptable), 
 					   confirm,
-					   1, 2, row, row+1);
-		row++;
+					   1, 2, row - 1, row);
 
 		/* Strength bar: */
+		gtk_table_resize (GTK_TABLE (ptable), ++row, 2);
 		strength_bar = gtk_progress_bar_new ();
 		gtk_progress_bar_set_text (GTK_PROGRESS_BAR (strength_bar), _("New password strength"));
 		g_signal_connect ((gpointer) entry, "changed",
@@ -327,13 +391,43 @@ run_dialog (const char *title,
 				  strength_bar);
 		gtk_table_attach_defaults (GTK_TABLE (ptable), 
 					   strength_bar,
-					   1, 2, row, row+1);
-		row++;
+					   1, 2, row - 1, row);
+	}
+	
+	check = NULL;
+	if (g_key_file_get_boolean (input_data, "check", "check-enable", NULL)) {
+		value = g_key_file_get_value (input_data, "check", "check-text", NULL);
+		if (!value)
+			fatal ("'check-enable' set, but no 'check-text'", NULL);
+		gtk_table_resize (GTK_TABLE (ptable), ++row, 2);
+		check = gtk_check_button_new_with_mnemonic (value);
+		gtk_table_attach_defaults (GTK_TABLE (ptable), check,  
+		                           1, 2, row - 1, row);
+		g_free (value);
 	}
 	
 	if (row > 0)
 		gtk_widget_show_all (ptable);
 
+	/* 
+	 * When passwords are involved we grab the keyboard so that people
+	 * don't accidentally type their passwords in other windows.
+	 */
+	if (include_password || include_confirm || include_original) { 
+		g_signal_connect (dialog, "map-event", G_CALLBACK (grab_keyboard), NULL);
+		g_signal_connect (dialog, "unmap-event", G_CALLBACK (ungrab_keyboard), NULL);
+		g_signal_connect (dialog, "window-state-event", G_CALLBACK (window_state_changed), NULL); 
+	}
+
+	/* 
+	 * We do this to guarantee the dialog comes up on top. Since the code that
+	 * that prompted this dialog is many processes away, we can't figure out 
+	 * a window to be transient for. 
+	 */
+	gtk_window_set_keep_above (GTK_WINDOW (dialog), TRUE);
+	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
+	gtk_window_set_type_hint (GTK_WINDOW (dialog), GDK_WINDOW_TYPE_HINT_NORMAL);
+	
 	/*
 	 * We do this as late as possible, so all the memory the process needs is 
 	 * allocated in memory. This prevents mapping failures.
@@ -377,6 +471,17 @@ run_dialog (const char *title,
 		}
 		*password_out = g_strdup (password);
 	}
+	
+	if (check != NULL && response >= GKR_ASK_RESPONSE_ALLOW) {
+		g_key_file_set_boolean (output_data, "check", "check-active", 
+		                        gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (check)));
+	}
+	
+	if (location != NULL) {
+		text = gkr_ask_tool_get_location (location);
+		if (text != NULL)
+			g_key_file_set_value (output_data, "location", "location-selected", text);
+	}		                      
 
 	gtk_widget_destroy (dialog);
 	
@@ -386,41 +491,42 @@ run_dialog (const char *title,
 static void 
 prepare_dialog (void)
 {
-	gchar *password, *original;
+	gchar *password = NULL;
+	gchar *original = NULL;
 	const gchar* buttons[6];
 	guint resps[6];
 	int i = 0;
 	guint response;
-	
-	g_assert (env_title);
-	g_assert (env_primary);
+	guint flags;
 	
 	memset (buttons, 0, sizeof (buttons));
 	memset (resps, 0, sizeof (resps));
 	
 	password = original = NULL;
 	
-	if (!(env_flags & GKR_ASK_REQUEST_BUTTONS_MASK)) 
-		env_flags |= GKR_ASK_REQUEST_OK_DENY_BUTTONS;
+	flags = g_key_file_get_integer (input_data, "general", "flags", NULL);
+	
+	if (!(flags & GKR_ASK_REQUEST_BUTTONS_MASK)) 
+		flags |= GKR_ASK_REQUEST_OK_DENY_BUTTONS;
 	
 	/* In order of preference for default response */
-	if (env_flags & GKR_ASK_REQUEST_DENY_BUTTON) {
+	if (flags & GKR_ASK_REQUEST_DENY_BUTTON) {
 		buttons[i] = _("Deny");
 		resps[i++] = GKR_ASK_RESPONSE_DENY;
 	}
-	if (env_flags & GKR_ASK_REQUEST_CANCEL_BUTTON) {
+	if (flags & GKR_ASK_REQUEST_CANCEL_BUTTON) {
 		buttons[i] = GTK_STOCK_CANCEL;
 		resps[i++] = GKR_ASK_RESPONSE_DENY;
 	}
-	if (env_flags & GKR_ASK_REQUEST_OK_BUTTON) {
+	if (flags & GKR_ASK_REQUEST_OK_BUTTON) {
 		buttons[i] = GTK_STOCK_OK;
 		resps[i++] = GKR_ASK_RESPONSE_ALLOW;
 	}
-	if (env_flags & GKR_ASK_REQUEST_ALLOW_BUTTON) {
+	if (flags & GKR_ASK_REQUEST_ALLOW_BUTTON) {
 		buttons[i] = _("Allow _Once");
 		resps[i++] = GKR_ASK_RESPONSE_ALLOW;
 	}
-	if (env_flags & GKR_ASK_REQUEST_ALLOW_FOREVER_BUTTON) {
+	if (flags & GKR_ASK_REQUEST_ALLOW_FOREVER_BUTTON) {
 		buttons[i] = _("_Always Allow");
 		resps[i++] = GKR_ASK_RESPONSE_ALLOW_FOREVER;
 	}
@@ -428,10 +534,9 @@ prepare_dialog (void)
 	g_assert (i > 0);
 	
 	password = NULL;
-	response = run_dialog (env_title, env_primary, env_secondary,
-	                       env_flags & GKR_ASK_REQUEST_PASSWORD, 
-	                       env_flags & GKR_ASK_REQUEST_CONFIRM_PASSWORD,
-	                       env_flags & GKR_ASK_REQUEST_ORIGINAL_PASSWORD,
+	response = run_dialog (flags & GKR_ASK_REQUEST_PASSWORD, 
+	                       flags & GKR_ASK_REQUEST_CONFIRM_PASSWORD,
+	                       flags & GKR_ASK_REQUEST_ORIGINAL_PASSWORD,
 	                       &password, &original, 
 	                       resps[i - 1], /* default response, last one added */
 	                       buttons[0], resps[0],
@@ -449,38 +554,84 @@ prepare_dialog (void)
 		response = GKR_ASK_RESPONSE_FAILURE;
 	}
 	
-	/* Send back the response */
-	printf ("%d\n", response);
-	
-	if (response >= GKR_ASK_RESPONSE_ALLOW) {
+	if (!password)
+		password = g_strdup ("");
+	if (!original)
+		original = g_strdup ("");
+
+	/* First two lines of the response are always the passwords */
+	if (response < GKR_ASK_RESPONSE_ALLOW || !(flags & GKR_ASK_REQUEST_PASSWORD))
+		password[0] = 0;
+	printf ("%s\n", password);
 		
-		/* Send back the password */
-		if ((env_flags & GKR_ASK_REQUEST_PASSWORD) && password)
-			printf ("%s\n", password);
+	if (response < GKR_ASK_RESPONSE_ALLOW || !(flags & GKR_ASK_REQUEST_ORIGINAL_PASSWORD))
+		original[0] = 0;
+	printf ("%s\n", original);
 	
-		/* And the original */
-		if ((env_flags & GKR_ASK_REQUEST_ORIGINAL_PASSWORD) && original)
-			printf ("%s", original);
+	fflush (stdout);
+	
+	/* Send back the response */
+	g_key_file_set_integer (output_data, "general", "response", response);
+	
+	g_free (password);
+	g_free (original);
+}
+
+static gchar*
+read_all_input (void)
+{
+	GString *data = g_string_new ("");
+	gchar buf[256];
+	int r;
+
+	for (;;) {
+		r = read (0, buf, sizeof (buf));
+		if (r < 0) {
+			if (errno == EAGAIN || errno == EINTR)
+				continue;
+			g_warning ("couldn't read dialog instructions from input: %s",
+			           g_strerror (errno));
+			exit (1);
+		} 
+		if (r == 0)
+			break;
+		g_string_append_len (data, buf, r);
+	}
+	
+	return g_string_free (data, FALSE);
+}
+
+static void
+write_all_output (const gchar *data, gsize len)
+{
+	int res;
+	
+	while (len > 0) {
+		res = write (1, data, len);
+		if (res <= 0) {
+			if (errno == EAGAIN && errno == EINTR)
+				continue;
+			g_warning ("couldn't write dialog response to output: %s",
+			           g_strerror (errno));
+			exit (1);
+		} else  {
+			len -= res;
+			data += res;
+		}
 	}
 }
 
 int
 main (int argc, char *argv[])
 {
-	const gchar *flags;
+	GError *err = NULL;
+	gchar *data;
+	gboolean ret;
+	gsize length;
 	
-	env_title = g_getenv ("ASK_TITLE");
-	env_primary = g_getenv ("ASK_PRIMARY");
-	env_secondary = g_getenv ("ASK_SECONDARY");
-	flags = g_getenv ("ASK_FLAGS");
+	input_data = g_key_file_new ();
+	output_data = g_key_file_new ();
 	
-	if (!env_title || !env_primary || !env_secondary || !flags) {
-		g_printerr (_("Must be run from gnome-keyring\n"));
-		return 1;
-	}
-	
-	env_flags = atoi (flags);
-
 	gtk_init (&argc, &argv);
 
 #ifdef HAVE_LOCALE_H
@@ -494,8 +645,30 @@ main (int argc, char *argv[])
 	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
 #endif
 
+	data = read_all_input ();
+	g_assert (data);
+	
+	if (!data[0])
+		fatal ("no dialog instructions", NULL);	
+	
+	ret = g_key_file_load_from_data (input_data, data, strlen (data), G_KEY_FILE_NONE, &err);
+	g_free (data);
+
+	if (!ret)
+		fatal ("couldn't parse dialog instructions", err ? err->message : "");
+
 	prepare_dialog ();
 	
+	g_key_file_free (input_data);
+	data = g_key_file_to_data (output_data, &length, &err);
+	g_key_file_free (output_data);
+	
+	if (!data)
+		fatal ("couldn't format dialog response: %s", err ? err->message : ""); 
+	
+	write_all_output (data, length);
+	g_free (data);
+	
 	return 0;
 }
 
